////////////////////////////////////////////////////////////////////////////////
// OLE Automation for XML with Delphi
// XMLAuto.pas
// fabrice.marguerie@cogisoft.fr
// http://www.cogisoft.fr/downloads
// 22/04/1999
////////////////////////////////////////////////////////////////////////////////

unit XMLAuto;

interface

uses
  ComObj,
  MSXML_TLB;

type
  TVarDispProc = procedure (Result: PVariant; const Instance: Variant;
      CallDesc: PCallDesc; Params: Pointer); cdecl;
  TXMLAutoProperty = (xpAutoCreateElement, xpEnableDefaultProperty, xpEnableDOM);
  TXMLAutoProperties = set of TXMLAutoProperty;

var
  OldVarDispProc: TVarDispProc = nil;


// functions to get or set values of XML Auto configuration variables
//  these functions are needed since threadvar cannot be in interface section if
//  used in a package
function GetXMLAutoProperty(XMLAutoProperty: TXMLAutoProperty): Boolean;
procedure SetXMLAutoProperty(XMLAutoProperty: TXMLAutoProperty; Value: Boolean);
procedure SetXMLAutoProperties(XMLAutoProperties: TXMLAutoProperties; Value: Boolean);


implementation


uses
  Windows, ActiveX, SysUtils;


threadvar
  AutoCreateElement: Boolean;
  EnableDefaultProperty: Boolean;
  EnableDOM: Boolean;

function GetXMLAutoProperty(XMLAutoProperty: TXMLAutoProperty): Boolean;
begin
  case XMLAutoProperty of
    xpAutoCreateElement: Result := AutoCreateElement;
    xpEnableDefaultProperty: Result := EnableDefaultProperty;
    xpEnableDOM: Result := EnableDOM;
    else Result := False;
  end;
end;

procedure SetXMLAutoProperty(XMLAutoProperty: TXMLAutoProperty; Value: Boolean);
begin
  case XMLAutoProperty of
    xpAutoCreateElement: AutoCreateElement := Value;
    xpEnableDefaultProperty: EnableDefaultProperty := Value;
    xpEnableDOM: EnableDOM := Value;
  end;
end;

procedure SetXMLAutoProperties(XMLAutoProperties: TXMLAutoProperties; Value: Boolean);
begin
  if xpAutoCreateElement in XMLAutoProperties then
    AutoCreateElement := Value;
  if xpEnableDefaultProperty in XMLAutoProperties then
    EnableDefaultProperty := Value;
  if xpEnableDOM in XMLAutoProperties then
    EnableDOM := Value;
end;

function GetIDsOfNames(const Dispatch: IDispatch; Names: PChar;
  NameCount: Integer; DispIDs: PDispIDList): HResult;
type
  PNamesArray = ^TNamesArray;
  TNamesArray = array[0..0] of PWideChar;
var
  N, SrcLen, DestLen: Integer;
  Src: PChar;
  Dest: PWideChar;
  NameRefs: PNamesArray;
  StackTop: Pointer;
begin
  Src := Names;
  N := 0;
  asm
    MOV  StackTop, ESP
    MOV  EAX, NameCount
    INC  EAX
    SHL  EAX, 2  // sizeof pointer = 4
    SUB  ESP, EAX
    LEA  EAX, NameRefs
    MOV  [EAX], ESP
  end;
  repeat
    SrcLen := StrLen(Src);
    DestLen := MultiByteToWideChar(0, 0, Src, SrcLen, nil, 0) + 1;
    asm
      MOV  EAX, DestLen
      ADD  EAX, EAX
      ADD  EAX, 3      // round up to 4 byte boundary
      AND  EAX, not 3
      SUB  ESP, EAX
      LEA  EAX, Dest
      MOV  [EAX], ESP
    end;
    if N = 0 then NameRefs[0] := Dest else NameRefs[NameCount - N] := Dest;
    MultiByteToWideChar(0, 0, Src, SrcLen, Dest, DestLen);
    Dest[DestLen-1] := #0;
    Inc(Src, SrcLen+1);
    Inc(N);
  until N = NameCount;
  Result := Dispatch.GetIDsOfNames(GUID_NULL, NameRefs, NameCount,
    GetThreadLocale, DispIDs);
  asm
    MOV  ESP, StackTop
  end;
end;

procedure VarDispInvoke(Result: PVariant; const Instance: Variant;
  CallDesc: PCallDesc; Params: Pointer); cdecl;
resourcestring
  STypeUnhandled = 'Type unhandled';
const
  atTypeMask = $7F;
  varStrArg = $0048;
  MaxDispArgs = 64;
var
  DispIDs: array[0..MaxDispArgs - 1] of Integer;
  Dispatch: IDispatch;
  FirstArgType, LastArgType: Integer;
  Index: Integer;
  LastParamPtr: Pointer;
  NameExists: HResult;
  Node, ResultNode: IXMLDOMNode;
  NodeName: string;
  OwnerDocument: IXMLDOMDocument;
  Param: string;

  procedure HandlePropertyPut;
  begin
    if CallDesc^.CallType = DISPATCH_PROPERTYPUT then
    begin
      case LastArgType of
        varStrArg: ResultNode.text := PString(LastParamPtr)^;
        varInteger: ResultNode.text := IntTostr(PInteger(LastParamPtr)^);
        else raise Exception.Create(STypeUnhandled);
      end;
    end;
  end;

  procedure CreateElement;
  begin
    if Node.QueryInterface(IXMLDOMDocument, OwnerDocument) <> S_OK then
      OwnerDocument := Node.ownerDocument;
    ResultNode := OwnerDocument.createElement(Param);
    Node.appendChild(ResultNode);
    HandlePropertyPut;
    if Result <> nil then
      Result^ := ResultNode;
  end;

begin
  // check Variant is IDispatch
  if TVarData(Instance).VType = varDispatch then
    Dispatch := IDispatch(TVarData(Instance).VDispatch)
  else if TVarData(Instance).VType = (varDispatch or varByRef) then
    Dispatch := IDispatch(TVarData(Instance).VPointer^)
  else
    raise EOleError.Create('VarNotObject');

  // IXMLDOMNode ?
  if Dispatch.QueryInterface(IXMLDOMNode, Node) <> S_OK then
  begin
    OldVarDispProc(Result, Instance, CallDesc, Params);
    Exit;
  end;

  // parameters
  Param := PChar(@CallDesc^.ArgTypes[CallDesc^.ArgCount]);
  FirstArgType := CallDesc^.ArgTypes[0] and atTypeMask;
  LastArgType := CallDesc^.ArgTypes[CallDesc^.ArgCount-1] and atTypeMask;
  LastParamPtr := Pointer(Integer(@Params) + (CallDesc^.ArgCount-1)*SizeOf(Pointer));

  // handle indices
  if FirstArgType = varInteger then
  begin
    Index := Integer(Params);
    NodeName := Param+'['+IntToStr(Index)+']';
  end
  else
    NodeName := Param;

  // element ?
  ResultNode := Node.selectSingleNode(NodeName);
  // attribute ?
  if ResultNode = nil then
    ResultNode := Node.selectSingleNode('@'+NodeName);

  if ResultNode <> nil then
  begin
    HandlePropertyPut;

    // result
    if Result <> nil then
    begin
      if EnableDefaultProperty and (ResultNode.firstChild <> nil) and
          (CompareText(ResultNode.firstChild.nodeName, '#text') = 0) then
        Result^ := ResultNode.text
      else
        Result^ := ResultNode;
    end;
    Exit;
  end;

  // default
  if EnableDOM then
  begin
    NameExists := GetIDsOfNames(Dispatch,
        PChar(@CallDesc^.ArgTypes[CallDesc^.ArgCount]),
        CallDesc^.NamedArgCount+1, @DispIDs);
    if NameExists = DISP_E_UNKNOWNNAME then
    begin
      if AutoCreateElement then
        CreateElement;
    end
    else
      OldVarDispProc(Result, Instance, CallDesc, Params);
  end
  else if EnableDefaultProperty and AutoCreateElement then
    CreateElement;
end;


initialization
  OldVarDispProc := VarDispProc;
  VarDispProc := @VarDispInvoke;

  AutoCreateElement := False;
  EnableDefaultProperty := False;
  EnableDOM := True;

finalization
  VarDispProc := @OldVarDispProc;
end.
